# -*- coding: utf-8 -*-
"""Python's built-in :mod:`functools` module builds several useful
utilities on top of Python's first-class function
support. ``funcutils`` generally stays in the same vein, adding to and
correcting Python's standard metaprogramming facilities.
"""
from __future__ import print_function

import sys
import re
import inspect
import functools
import itertools
from types import MethodType, FunctionType

try:
    xrange
    make_method = MethodType
except NameError:
    # Python 3
    make_method = lambda desc, obj, obj_type: MethodType(desc, obj)
    basestring = (str, bytes)  # Python 3 compat
    _IS_PY2 = False
else:
    _IS_PY2 = True


def formatannotation(annotation, base_module=None):
    return repr(annotation)


def get_module_callables(mod, ignore=None):
    """Returns two maps of (*types*, *funcs*) from *mod*, optionally
    ignoring based on the :class:`bool` return value of the *ignore*
    callable. *mod* can be a string name of a module in
    :data:`sys.modules` or the module instance itself.
    """
    if isinstance(mod, basestring):
        mod = sys.modules[mod]
    types, funcs = {}, {}
    for attr_name in dir(mod):
        if ignore and ignore(attr_name):
            continue
        try:
            attr = getattr(mod, attr_name)
        except Exception:
            continue
        try:
            attr_mod_name = attr.__module__
        except AttributeError:
            continue
        if attr_mod_name != mod.__name__:
            continue
        if isinstance(attr, type):
            types[attr_name] = attr
        elif callable(attr):
            funcs[attr_name] = attr
    return types, funcs


def mro_items(type_obj):
    """Takes a type and returns an iterator over all class variables
    throughout the type hierarchy (respecting the MRO).

    >>> sorted(set([k for k, v in mro_items(int) if not k.startswith('__') and 'bytes' not in k and not callable(v)]))
    ['denominator', 'imag', 'numerator', 'real']
    """
    # TODO: handle slots?
    return itertools.chain.from_iterable(ct.__dict__.items()
                                         for ct in type_obj.__mro__)


def dir_dict(obj, raise_exc=False):
    """Return a dictionary of attribute names to values for a given
    object. Unlike ``obj.__dict__``, this function returns all
    attributes on the object, including ones on parent classes.
    """
    # TODO: separate function for handling descriptors on types?
    ret = {}
    for k in dir(obj):
        try:
            ret[k] = getattr(obj, k)
        except Exception:
            if raise_exc:
                raise
    return ret


def copy_function(orig, copy_dict=True):
    """Returns a shallow copy of the function, including code object,
    globals, closure, etc.

    >>> func = lambda: func
    >>> func() is func
    True
    >>> func_copy = copy_function(func)
    >>> func_copy() is func
    True
    >>> func_copy is not func
    True

    Args:
        orig (function): The function to be copied. Must be a
            function, not just any method or callable.
        copy_dict (bool): Also copy any attributes set on the function
            instance. Defaults to ``True``.
    """
    ret = FunctionType(orig.__code__,
                       orig.__globals__,
                       name=orig.__name__,
                       argdefs=getattr(orig, "__defaults__", None),
                       closure=getattr(orig, "__closure__", None))
    if copy_dict:
        ret.__dict__.update(orig.__dict__)
    return ret


def partial_ordering(cls):
    """Class decorator, similar to :func:`functools.total_ordering`,
    except it is used to define `partial orderings`_ (i.e., it is
    possible that *x* is neither greater than, equal to, or less than
    *y*). It assumes the presence of the ``__le__()`` and ``__ge__()``
    method, but nothing else. It will not override any existing
    additional comparison methods.

    .. _partial orderings: https://en.wikipedia.org/wiki/Partially_ordered_set

    >>> @partial_ordering
    ... class MySet(set):
    ...     def __le__(self, other):
    ...         return self.issubset(other)
    ...     def __ge__(self, other):
    ...         return self.issuperset(other)
    ...
    >>> a = MySet([1,2,3])
    >>> b = MySet([1,2])
    >>> c = MySet([1,2,4])
    >>> b < a
    True
    >>> b > a
    False
    >>> b < c
    True
    >>> a < c
    False
    >>> c > a
    False
    """
    def __lt__(self, other): return self <= other and not self >= other
    def __gt__(self, other): return self >= other and not self <= other
    def __eq__(self, other): return self >= other and self <= other

    if not hasattr(cls, '__lt__'): cls.__lt__ = __lt__
    if not hasattr(cls, '__gt__'): cls.__gt__ = __gt__
    if not hasattr(cls, '__eq__'): cls.__eq__ = __eq__

    return cls


class InstancePartial(functools.partial):
    """:class:`functools.partial` is a huge convenience for anyone
    working with Python's great first-class functions. It allows
    developers to curry arguments and incrementally create simpler
    callables for a variety of use cases.

    Unfortunately there's one big gap in its usefulness:
    methods. Partials just don't get bound as methods and
    automatically handed a reference to ``self``. The
    ``InstancePartial`` type remedies this by inheriting from
    :class:`functools.partial` and implementing the necessary
    descriptor protocol. There are no other differences in
    implementation or usage. :class:`CachedInstancePartial`, below,
    has the same ability, but is slightly more efficient.

    """
    def __get__(self, obj, obj_type):
        return make_method(self, obj, obj_type)


class CachedInstancePartial(functools.partial):
    """The ``CachedInstancePartial`` is virtually the same as
    :class:`InstancePartial`, adding support for method-usage to
    :class:`functools.partial`, except that upon first access, it
    caches the bound method on the associated object, speeding it up
    for future accesses, and bringing the method call overhead to
    about the same as non-``partial`` methods.

    See the :class:`InstancePartial` docstring for more details.
    """
    def __get__(self, obj, obj_type):
        # These assignments could've been in __init__, but there was
        # no simple way to do it without breaking one of PyPy or Py3.
        self.__name__ = None
        self.__doc__ = self.func.__doc__
        self.__module__ = self.func.__module__

        name = self.__name__
        if name is None:
            for k, v in mro_items(obj_type):
                if v is self:
                    self.__name__ = name = k
        if obj is None:
            return make_method(self, obj, obj_type)
        try:
            # since this is a data descriptor, this block
            # is probably only hit once (per object)
            return obj.__dict__[name]
        except KeyError:
            obj.__dict__[name] = ret = make_method(self, obj, obj_type)
            return ret

partial = CachedInstancePartial


# # #
# # # Function builder
# # #


def wraps(func, injected=None, **kw):
    """Modeled after the built-in :func:`functools.wraps`, this function is
    used to make your decorator's wrapper functions reflect the
    wrapped function's:

      * Name
      * Documentation
      * Module
      * Signature

    The built-in :func:`functools.wraps` copies the first three, but
    does not copy the signature. This version of ``wraps`` can copy
    the inner function's signature exactly, allowing seamless usage
    and :mod:`introspection <inspect>`. Usage is identical to the
    built-in version::

        >>> from boltons.funcutils import wraps
        >>>
        >>> def print_return(func):
        ...     @wraps(func)
        ...     def wrapper(*args, **kwargs):
        ...         ret = func(*args, **kwargs)
        ...         print(ret)
        ...         return ret
        ...     return wrapper
        ...
        >>> @print_return
        ... def example():
        ...     '''docstring'''
        ...     return 'example return value'
        >>>
        >>> val = example()
        example return value
        >>> example.__name__
        'example'
        >>> example.__doc__
        'docstring'

    In addition, the boltons version of wraps supports modifying the
    outer signature based on the inner signature. By passing a list of
    *injected* argument names, those arguments will be removed from
    the outer wrapper's signature, allowing your decorator to provide
    arguments that aren't passed in.

    Args:

        func (function): The callable whose attributes are to be copied.
        injected (list): An optional list of argument names which
            should not appear in the new wrapper's signature.
        update_dict (bool): Whether to copy other, non-standard
            attributes of *func* over to the wrapper. Defaults to True.
        inject_to_varkw (bool): Ignore missing arguments when a
            ``**kwargs``-type catch-all is present. Defaults to True.

    For more in-depth wrapping of functions, see the
    :class:`FunctionBuilder` type, on which wraps was built.

    """
    # TODO: maybe automatically use normal wraps in the very rare case
    # that the signatures actually match and no adapter is needed.
    if injected is None:
        injected = []
    elif isinstance(injected, basestring):
        injected = [injected]
    else:
        injected = list(injected)

    if isinstance(func, (classmethod, staticmethod)):
        raise TypeError('wraps does not support wrapping classmethods and'
                        ' staticmethods, change the order of wrapping to'
                        ' wrap the underlying function: %r'
                        % (getattr(func, '__func__', None),))

    update_dict = kw.pop('update_dict', True)
    inject_to_varkw = kw.pop('inject_to_varkw', True)
    if kw:
        raise TypeError('unexpected kwargs: %r' % kw.keys())

    fb = FunctionBuilder.from_func(func)
    for arg in injected:
        try:
            fb.remove_arg(arg)
        except MissingArgument:
            if inject_to_varkw and fb.varkw is not None:
                continue  # keyword arg will be caught by the varkw
            raise

    fb.body = 'return _call(%s)' % fb.get_invocation_str()

    def wrapper_wrapper(wrapper_func):
        execdict = dict(_call=wrapper_func, _func=func)
        fully_wrapped = fb.get_func(execdict, with_dict=update_dict)
        fully_wrapped.__wrapped__ = func  # ref to the original function (#115)

        return fully_wrapped

    return wrapper_wrapper


class FunctionBuilder(object):
    """The FunctionBuilder type provides an interface for programmatically
    creating new functions, either based on existing functions or from
    scratch.

    Values are passed in at construction or set as attributes on the
    instance. For creating a new function based of an existing one,
    see the :meth:`~FunctionBuilder.from_func` classmethod. At any
    point, :meth:`~FunctionBuilder.get_func` can be called to get a
    newly compiled function, based on the values configured.

    >>> fb = FunctionBuilder('return_five', doc='returns the integer 5',
    ...                      body='return 5')
    >>> f = fb.get_func()
    >>> f()
    5
    >>> fb.varkw = 'kw'
    >>> f_kw = fb.get_func()
    >>> f_kw(ignored_arg='ignored_val')
    5

    Note that function signatures themselves changed quite a bit in
    Python 3, so several arguments are only applicable to
    FunctionBuilder in Python 3. Except for *name*, all arguments to
    the constructor are keyword arguments.

    Args:
        name (str): Name of the function.
        doc (str): `Docstring`_ for the function, defaults to empty.
        module (str): Name of the module from which this function was
            imported. Defaults to None.
        body (str): String version of the code representing the body
            of the function. Defaults to ``'pass'``, which will result
            in a function which does nothing and returns ``None``.
        args (list): List of argument names, defaults to empty list,
            denoting no arguments.
        varargs (str): Name of the catch-all variable for positional
            arguments. E.g., "args" if the resultant function is to have
            ``*args`` in the signature. Defaults to None.
        varkw (str): Name of the catch-all variable for keyword
            arguments. E.g., "kwargs" if the resultant function is to have
            ``**kwargs`` in the signature. Defaults to None.
        defaults (dict): A mapping of argument names to default values.
        kwonlyargs (list): Argument names which are only valid as
            keyword arguments. **Python 3 only.**
        kwonlydefaults (dict): A mapping, same as normal *defaults*,
            but only for the *kwonlyargs*. **Python 3 only.**
        annotations (dict): Mapping of type hints and so
            forth. **Python 3 only.**
        filename (str): The filename that will appear in
            tracebacks. Defaults to "boltons.funcutils.FunctionBuilder".
        indent (int): Number of spaces with which to indent the
            function *body*. Values less than 1 will result in an error.
        dict (dict): Any other attributes which should be added to the
            functions compiled with this FunctionBuilder.

    All of these arguments are also made available as attributes which
    can be mutated as necessary.

    .. _Docstring: https://en.wikipedia.org/wiki/Docstring#Python

    """

    if _IS_PY2:
        _argspec_defaults = {'args': list,
                             'varargs': lambda: None,
                             'varkw': lambda: None,
                             'defaults': lambda: None}

        @classmethod
        def _argspec_to_dict(cls, f):
            args, varargs, varkw, defaults = inspect.getargspec(f)
            return {'args': args,
                    'varargs': varargs,
                    'varkw': varkw,
                    'defaults': defaults}

    else:
        _argspec_defaults = {'args': list,
                             'varargs': lambda: None,
                             'varkw': lambda: None,
                             'defaults': lambda: None,
                             'kwonlyargs': list,
                             'kwonlydefaults': dict,
                             'annotations': dict}

        @classmethod
        def _argspec_to_dict(cls, f):
            argspec = inspect.getfullargspec(f)
            return dict((attr, getattr(argspec, attr))
                        for attr in cls._argspec_defaults)

    _defaults = {'doc': str,
                 'dict': dict,
                 'module': lambda: None,
                 'body': lambda: 'pass',
                 'indent': lambda: 4,
                 'lineno': lambda: 1,
                 'filename': lambda: 'boltons.funcutils.FunctionBuilder'}

    _defaults.update(_argspec_defaults)

    _compile_count = itertools.count()

    def __init__(self, name, **kw):
        self.name = name
        for a, default_factory in self._defaults.items():
            val = kw.pop(a, None)
            if val is None:
                val = default_factory()
            setattr(self, a, val)

        if kw:
            raise TypeError('unexpected kwargs: %r' % kw.keys())
        return

    # def get_argspec(self):  # TODO

    if _IS_PY2:
        def get_sig_str(self):
            return inspect.formatargspec(self.args, self.varargs,
                                         self.varkw, [])

        def get_invocation_str(self):
            return inspect.formatargspec(self.args, self.varargs,
                                         self.varkw, [])[1:-1]
    else:
        def get_sig_str(self, formatannotation=formatannotation):
            return inspect.formatargspec(self.args,
                                         self.varargs,
                                         self.varkw,
                                         [],
                                         self.kwonlyargs,
                                         {},
                                         self.annotations,
                                         formatannotation=formatannotation
                                         )

        _KWONLY_MARKER = re.compile(r"""
        \*     # a star
        \s*    # followed by any amount of whitespace
        ,      # followed by a comma
        \s*    # followed by any amount of whitespace
        """, re.VERBOSE)

        def get_invocation_str(self):
            kwonly_pairs = None
            formatters = {}
            if self.kwonlyargs:
                kwonly_pairs = dict((arg, arg)
                                    for arg in self.kwonlyargs)
                formatters['formatvalue'] = lambda value: '=' + value

            sig = inspect.formatargspec(self.args,
                                        self.varargs,
                                        self.varkw,
                                        [],
                                        kwonly_pairs,
                                        kwonly_pairs,
                                        {},
                                        **formatters)
            sig = self._KWONLY_MARKER.sub('', sig)
            return sig[1:-1]

    @classmethod
    def from_func(cls, func):
        """Create a new FunctionBuilder instance based on an existing
        function. The original function will not be stored or
        modified.
        """
        # TODO: copy_body? gonna need a good signature regex.
        # TODO: might worry about __closure__?
        if not callable(func):
            raise TypeError('expected callable object, not %r' % (func,))

        kwargs = {'name': func.__name__,
                  'doc': func.__doc__,
                  'filename': func.__code__.co_filename,
                  'lineno': func.__code__.co_firstlineno,
                  'module': func.__module__,
                  'dict': getattr(func, '__dict__', {})}

        kwargs.update(cls._argspec_to_dict(func))

        return cls(**kwargs)

    def get_func(self, execdict=None, add_source=True, with_dict=True, formatannotation=formatannotation):
        """Compile and return a new function based on the current values of
        the FunctionBuilder.

        Args:
            execdict (dict): The dictionary representing the scope in
                which the compilation should take place. Defaults to an empty
                dict.
            add_source (bool): Whether to add the source used to a
                special ``__source__`` attribute on the resulting
                function. Defaults to True.
            with_dict (bool): Add any custom attributes, if
                applicable. Defaults to True.

        To see an example of usage, see the implementation of
        :func:`~boltons.funcutils.wraps`.
        """
        execdict = execdict or {}
        body = self.body or self._default_body

        tmpl = 'def {name}{sig_str}:'
        if self.doc:
            tmpl += '\n    """{doc}"""'
        tmpl += '\n{body}'

        body = _indent(self.body, ' ' * self.indent)

        name = self.name.replace('<', '_').replace('>', '_')  # lambdas
        src = tmpl.format(name=name, sig_str=self.get_sig_str(formatannotation),
                          doc=self.doc, body=body)
        self._compile(src, execdict)
        func = execdict[name]

        func.__name__ = self.name
        func.__doc__ = self.doc
        func.__defaults__ = self.defaults
        if not _IS_PY2:
            func.__kwdefaults__ = self.kwonlydefaults

        if with_dict:
            func.__dict__.update(self.dict)
        func.__module__ = self.module
        # TODO: caller module fallback?

        if add_source:
            func.__source__ = src

        # from types import CodeType
        # co = func.__code__
        # func.__code__ = CodeType(
        #     co.co_argcount, co.co_kwonlyargcount, co.co_nlocals,
        #     co.co_stacksize,
        #     co.co_flags,
        #     co.co_code,
        #     co.co_consts, co.co_names, co.co_varnames, co.co_filename,
        #     co.co_name, self.lineno, co.co_lnotab, co.co_freevars,
        #     co.co_cellvars)

        return func

    def get_defaults_dict(self):
        """Get a dictionary of function arguments with defaults and the
        respective values.
        """
        ret = dict(reversed(list(zip(reversed(self.args),
                                     reversed(self.defaults or [])))))
        return ret

    def remove_arg(self, arg_name):
        """Remove an argument from this FunctionBuilder's argument list. The
        resulting function will have one less argument per call to
        this function.

        Args:
            arg_name (str): The name of the argument to remove.

        Raises a :exc:`ValueError` if the argument is not present.

        """
        args = self.args
        d_dict = self.get_defaults_dict()
        try:
            args.remove(arg_name)
        except ValueError:
            try:
                self.kwonlyargs.remove(arg_name)
            except (AttributeError, ValueError):
                # py2, or py3 and missing from both
                exc = MissingArgument('arg %r not found in %s argument list:'
                                      ' %r' % (arg_name, self.name, args))
                exc.arg_name = arg_name
                raise exc
            else:
                self.kwonlydefaults.pop(arg_name, None)
        else:
            d_dict.pop(arg_name, None)
            self.defaults = tuple([d_dict[a] for a in args if a in d_dict])
        return

    def _compile(self, src, execdict):

        # filename = ('<%s-%d>'
        #             % (self.filename, next(self._compile_count),))
        try:
            code = compile(src, self.filename, 'single')
            exec(code, execdict)
        except Exception:
            raise
        return execdict


class MissingArgument(ValueError):
    pass


def _indent(text, margin, newline='\n', key=bool):
    "based on boltons.strutils.indent"
    indented_lines = [(margin + line if key(line) else line)
                      for line in text.splitlines()]
    return newline.join(indented_lines)


try:
    from functools import total_ordering  # 2.7+
except ImportError:
    # python 2.6
    def total_ordering(cls):
        """Class decorator that fills in missing comparators/ordering
        methods. Backport of :func:`functools.total_ordering` to work
        with Python 2.6.

        Code from http://code.activestate.com/recipes/576685/
        """
        convert = {
            '__lt__': [
                ('__gt__',
                 lambda self, other: not (self < other or self == other)),
                ('__le__',
                 lambda self, other: self < other or self == other),
                ('__ge__',
                 lambda self, other: not self < other)],
            '__le__': [
                ('__ge__',
                 lambda self, other: not self <= other or self == other),
                ('__lt__',
                 lambda self, other: self <= other and not self == other),
                ('__gt__',
                 lambda self, other: not self <= other)],
            '__gt__': [
                ('__lt__',
                 lambda self, other: not (self > other or self == other)),
                ('__ge__',
                 lambda self, other: self > other or self == other),
                ('__le__',
                 lambda self, other: not self > other)],
            '__ge__': [
                ('__le__',
                 lambda self, other: (not self >= other) or self == other),
                ('__gt__',
                 lambda self, other: self >= other and not self == other),
                ('__lt__',
                 lambda self, other: not self >= other)]
        }
        roots = set(dir(cls)) & set(convert)
        if not roots:
            raise ValueError('must define at least one ordering operation:'
                             ' < > <= >=')
        root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
        for opname, opfunc in convert[root]:
            if opname not in roots:
                opfunc.__name__ = opname
                opfunc.__doc__ = getattr(int, opname).__doc__
                setattr(cls, opname, opfunc)
        return cls

# end funcutils.py



from pygears.core.gear import gear
from pygears.typing import Uint
from pygears.core.intf import IntfOperPlugin


@gear
def lt(*din,
       din0_signed=b'typeof(din0, Int)',
       din1_signed=b'typeof(din1, Int)') -> Uint[1]:
    pass


class MulIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__lt__'] = lt



from pygears import gear, alternative, module
from pygears.typing import Array, Tuple, Uint, Queue
from pygears.util.utils import quiter_async, quiter


@gear
async def serialize(din: Array) -> b'din.dtype':
    async with din as val:
        for i in range(len(val)):
            yield val[i]


@alternative(serialize)
@gear
async def serialize_keep(
        din: Queue[Tuple[Array['dtype', 'width'], Uint['width']]]
) -> Queue['dtype']:

    async with din as qdata:
        data, active = qdata[0]

        active_out = [data[i] for i in range(len(data)) if active[i]]

        for d, last in quiter(active_out):
            yield module().tout((d, last and qdata.last))



import collections
import copy

from pygears.registry import registry
from pygears.typing.base import GenericMeta, param_subs, TypingMeta

from .type_match import TypeMatchError, type_match


def is_type_iterable(t):
    return (not isinstance(t, (str, bytes))) and isinstance(
        t, collections.Iterable)


def _copy_field_names(t, pat):
    if isinstance(t, GenericMeta) and isinstance(
            pat, GenericMeta) and (issubclass(t.base, pat.base)):
        for ta, pa in zip(t.args, pat.args):
            _copy_field_names(ta, pa)

        if hasattr(pat, '__parameters__'):
            t.__parameters__ = pat.__parameters__


def copy_field_names(t, pat):
    t = copy.copy(t)
    _copy_field_names(t, pat)
    return t


def type_is_specified(t):
    try:
        return t.is_specified()
    except Exception as e:
        if t is None:
            return True
        elif is_type_iterable(t):
            return all(type_is_specified(subt) for subt in t)
        else:
            return False


def resolve_param(val, match, namespace):

    is_templated_type = (isinstance(val, GenericMeta)
                         and (not val.is_generic()))

    if ((is_templated_type or is_type_iterable(val))
            and (not type_is_specified(val))):
        new_p = param_subs(val, match, namespace)
        if repr(new_p) != repr(val):
            return True, new_p

    elif isinstance(val, bytes):
        return True, eval(val, namespace, match)

    return False, None


def infer_ftypes(params, args, namespace={}, allow_incomplete=False):

    # Add all registered objects (types and transformations) to the namespace
    namespace = dict(namespace)
    namespace.update(registry('TypeArithNamespace'))

    def is_postponed(name, val):
        if isinstance(val, bytes):
            return True
        if (name in args):
            return True

        if (name == 'return'):
            return not type_is_specified(val)

        return False

    postponed = {
        name: val
        for name, val in params.items() if is_postponed(name, val)
    }
    match = {
        name: val
        for name, val in params.items() if name not in postponed
    }

    # print('Postponed: ', postponed)
    # print('Match: ', match)

    substituted = True
    final_check = False
    # Allow for keyword argument values to be templates and provide
    # a mechanism to resolve these template arguments
    while substituted or final_check:
        substituted = False
        # Loops until none of the parameters has been additionally resolved
        for name, val in postponed.copy().items():
            if name in args:
                try:
                    templ = val
                    if isinstance(val, bytes):
                        templ = templ.decode()

                    match.update(
                        type_match(
                            args[name],
                            templ,
                            match,
                            allow_incomplete=(not final_check)))
                except Exception as e:
                    raise TypeMatchError(
                        f"{str(e)}\n - when deducing type for argument "
                        f"'{name}'")
            try:
                substituted, new_p = resolve_param(val, match, namespace)
                if name in args:
                    new_p = args[name]
                    substituted = type_is_specified(new_p)

                if substituted:
                    if name in args:
                        new_p = copy_field_names(new_p, params[name])

                    match[name] = new_p
                    del postponed[name]
                    break
            except Exception as e:
                if final_check:
                    raise type(e)(f'{str(e)} - when resolving '
                                  f'parameter {name}: {val}')

        final_check = not substituted and not final_check

    # print('Final postponed: ', postponed)
    # print('Final match: ', match)

    for name, val in args.items():
        type_match(val, match[name], {})

    if postponed:
        name, value = next(iter(postponed.items()))
        print(match)
        raise TypeMatchError(f'Parameter "{name}" unresolved: {value}')

    return match



from pygears.typing import Queue, Tuple, typeof
from .flatten import flatten
from pygears import gear


def permute(dout, indices):
    dtype = dout.dtype
    lvl = 0
    if typeof(dtype, Queue):
        lvl = dtype.lvl
        dtype = dtype[0]
        # Turn Queue to Tuple for easier reordering, later we cast back to a
        # Queue
        dout = dout | Tuple | flatten

    reorder_indices = [0] * len(indices)
    for i, index in enumerate(indices):
        reorder_indices[index] = i

    # Cast back to a Queue at the end, if lvl=0, Tuple is returned
    out_type = Queue[dtype[tuple(reorder_indices)], lvl]

    # leave Queue delimiters in same place after reordering
    reorder_indices.append(-1)

    return dout[tuple(reorder_indices)] | out_type


def intf_arrange(*din, indices):
    return [din[i] for i in indices]


def tpl_arrange(*din, f, indices):
    return din[tuple(indices)]


@gear
def permuted_apply(*din, f, indices):
    if len(din) == len(indices):
        din_arranged = intf_arrange(*din, indices=indices)
    else:
        din_arranged = tpl_arrange(din[0], indices=indices)

    return permute(f(*din_arranged), indices)



import os
import traceback
import logging
import socket
import sys
import time
from http.server import HTTPServer, BaseHTTPRequestHandler
from http.client import HTTPConnection
from pygears import registry
from functools import partial
from traceback import extract_tb, format_list, walk_tb


def finish():
    os.system(f'pywave "exit"')


def pywave_running():
    return not os.system('pywave -c ""')


def pywave_reload():
    os.system(f'pywave "gtkwave::reLoadFile"')


def pywave_load_file(fn):
    if os.path.isfile(fn):
        os.system(f'pywave "gtkwave::loadFile {fn}"')


def ActivityWaveFactory(cfg):
    class ActivityWave(BaseHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super(ActivityWave, self).__init__(*args, **kwargs)

        def log_request(code='-', size='-'):
            pass

        def log_error(fmt, *args):
            pass

        def log_message(fmt, *args):
            pass

        def do_GET(self):
            """Respond to a GET request."""

            logging.info(f'Request received: {self.path}')

            try:
                self.send_response(200)
                if not self.path:
                    logging.info(f'Here?')
                    return
                elif self.path[0] == ':':
                    logging.info(f'Filedir received: {self.path}')
                    running = pywave_running()
                    if ('filedir' in cfg and cfg['filedir'] == self.path[1:]
                            and running):
                        pywave_reload()
                    else:
                        cfg['filedir'] = self.path[1:]
                        if running:
                            finish()

                        pywave_load_file(f"{cfg['filedir']}/pygears.vcd")
                        pywave_load_file(f"{cfg['filedir']}/issue.sav")

                elif self.path[0] == '/':
                    gear = self.path[1:]
                    if 'filedir' in cfg:
                        logging.info(f'Gear received: {self.path}')
                        pywave_load_file(f"{cfg['filedir']}/{gear}.sav")

                    logging.info(f'Here 0')
                    self.send_header("Content-type", "text/html")
                    self.end_headers()
                    self.wfile.write(
                        f"<html><head><title>{gear}.</title></head>".encode())

                    if 'filedir' in cfg:
                        self.wfile.write(
                            f"<body><p>Reading from: {cfg['filedir']}.</p>".
                            encode())

                    self.wfile.write(b"<p>")

                    fn = f"{cfg['filedir']}/{gear}.txt"
                    if os.path.isfile(fn):
                        with open(fn, 'r') as f:
                            for line in f:
                                self.wfile.write(f'{line}<br/>'.encode())

                    self.wfile.write(b"</p></body></html>")

            except Exception as e:
                type, value, tr = sys.exc_info()
                for s in format_list(extract_tb(tr)):
                    logging.info(s)

                logging.info(f'Error: {e}')

                raise e

    return ActivityWave


def restart_wave(sim, outdir, address):

    # conn = HTTPConnection(f'http://{address[0]}:{address[1]}')
    # conn.request('HEAD', f'http://{address[0]}:{address[1]}:{outdir}')
    # return

    while True:
        try:
            time.sleep(1)
            conn = HTTPConnection(f'{address[0]}:{address[1]}')
            conn.request('GET', f':{outdir}')
            return
        except ConnectionRefusedError:
            # print("Fail!")
            # raise e
            # return
            conn.close()
            command = [
                'python',
                os.path.join(os.path.dirname(__file__), 'activity_wave.py')
            ]

            pid = os.fork()
            if pid == 0:
                os.setsid()
                os.umask(0)
                os.execv(sys.executable, command)
            else:
                pass


def activity_wave(top):

    sim = registry('Simulator')
    outdir = os.path.abspath(registry('SimArtifactDir'))
    sim.events['at_exit'].append(
        partial(restart_wave, outdir=outdir, address=('localhost', 5000)))


def run(address=('', 5000), server_class=HTTPServer):
    cfg = {'address': address}

    logging.basicConfig(
        filename=os.path.expanduser(f'~/activity_wave_{address[1]}.log'),
        level=getattr(logging, 'INFO'),
        format='%(asctime)s %(levelname)-8s %(message)s')

    logging.info(f'ActivityWave process started, pid={os.getpid()}')
    httpd = server_class(address, ActivityWaveFactory(cfg))
    httpd.serve_forever()


if __name__ == "__main__":
    run()



from pygears.core.gear import gear
from pygears.typing import Union
from pygears import module
from .union import union_collapse


def mux_type(dtypes):
    return Union[dtypes]


@gear
async def mux(ctrl, *din) -> b'mux_type(din)':
    async with ctrl as c:
        async with din[c] as d:
            yield module().tout((d, c))


@gear
def mux_zip(ctrl, *din) -> b'mux_type(din)':
    pass


@gear
def mux_valve(ctrl, *din) -> b'mux_type(din)':
    pass


# @gear
# def mux_by(ctrl, *din, fmux=mux):
#     return fmux(ctrl, *din) | union_collapse



from pygears import registry
import atexit


class SimExtend:
    def __init__(self, top=None):
        self.sim = registry('Simulator')
        for name, event in self.sim.events.items():
            try:
                event.append(getattr(self, name))
            except AttributeError:
                pass

        try:
            atexit.register(self.at_exit, sim=None)
        except AttributeError:
            pass



from pygears.svgen.svmod import SVModuleGen
from pygears.svgen.inst import SVGenInstPlugin
from pygears import gear
from pygears.typing import Queue, Uint


@gear
async def qcnt(din: Queue, *, lvl=1, w_out=16) -> Queue[Uint['w_out']]:
    val = (0, ) * din.dtype.lvl
    cnt = 0
    last_el = False
    while not last_el:
        async with din as val:
            last_el = all(v for v in val[1:])
            cnt_mode = True if (lvl == din.dtype.lvl) else all(
                v for v in val[1:din.dtype.lvl - lvl + 1])
            if cnt_mode:
                yield (cnt, last_el)
                cnt += 1


class SVGenQcnt(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'lvl': self.node.params['lvl']
        }
        return template_env.render_local(__file__, "qcnt.j2", context)


class SVGenQcntPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][qcnt] = SVGenQcnt



import os
from collections import OrderedDict

from pygears import registry
from pygears.definitions import COMMON_SVLIB_DIR, COOKBOOK_SVLIB_DIR
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svparse import parse

from .inst import svgen_log


def find_in_dirs(fn, dirs):
    for d in dirs:
        full_path = os.path.join(d, fn)
        if os.path.exists(full_path):
            return full_path
    else:
        return None


class SVModuleGen:
    def __init__(self, node):
        self.node = node
        self.svgen_map = registry("SVGenMap")
        self._sv_module_name = None
        self.sv_module_path = None
        self.sv_params = {}

        if not self.is_generated:
            try:
                self.sv_module_path, self._sv_module_name, self.sv_params = self.get_sv_module_info(
                )
            except FileNotFoundError:
                svgen_log().warning(
                    f'SystemVerilog file not found for {self.node.name}')
        elif self.is_hierarchical:
            if find_in_dirs(self.sv_file_name,
                            registry('SVGenSystemVerilogPaths')):
                self.sv_module_path = None
                self._sv_module_name = self.hier_sv_path_name + '_hier'

    @property
    def is_generated(self):
        return self.is_hierarchical

    @property
    def is_hierarchical(self):
        return self.node.is_hierarchical

    @property
    def sv_module_name(self):
        if self.is_generated:
            return self._sv_module_name or self.hier_sv_path_name
        else:
            return self._sv_module_name or self.node.gear.definition.__name__

    @property
    def sv_inst_name(self):
        return f'{self.node.basename}_i'

    @property
    def sv_file_name(self):
        svgen_params = self.node.params.get('svgen', {})
        return svgen_params.get('svmod_fn', self.sv_module_name + ".sv")

    def get_sv_module_info(self):
        svmod_fn = self.sv_file_name
        if svmod_fn:
            svmod_path = find_in_dirs(svmod_fn,
                                      registry('SVGenSystemVerilogPaths'))
            if svmod_path:
                with open(svmod_path, 'r') as f:
                    name, _, _, svparams = parse(f.read())

                return svmod_path, name, svparams

        raise FileNotFoundError(
            f'SystemVerilog file not found for {self.node.name}')

    @property
    def params(self):
        return {
            k.upper(): int(v)
            for k, v in self.node.params.items()
            if (k.upper() in self.sv_params) and (
                int(v) != int(self.sv_params[k.upper()]['val']))
        }

    def sv_port_configs(self):
        for p in self.node.in_ports:
            yield self.get_sv_port_config(
                'consumer', type_=p.dtype, name=p.basename)

        for p in self.node.out_ports:
            yield self.get_sv_port_config(
                'producer', type_=p.dtype, name=p.basename)

    def get_sv_port_config(self, modport, type_, name):
        return {
            'modport': modport,
            'name': name,
            'type': type_,
            'width': int(type_),
            'local_type': type_
        }

    def get_synth_wrap(self, template_env):

        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'param_map': self.params
        }
        return template_env.render_local(__file__, "module_synth_wrap.j2",
                                         context)

    def get_out_port_map_intf_name(self, port):
        return self.svgen_map[port.consumer].basename

    def get_in_port_map_intf_name(self, port):
        intf = port.producer
        svgen_intf = self.svgen_map[intf]

        if len(intf.consumers) == 1:
            return svgen_intf.outname
        else:
            i = intf.consumers.index(port)
            return f'{svgen_intf.outname}[{i}]'

    @property
    def has_local_rst(self):
        return any(child.gear.definition.__name__ == 'local_rst'
                   for child in self.node.local_modules())

    @property
    def hier_sv_path_name(self):
        trimmed_name = self.node.name

        if trimmed_name.startswith('/'):
            trimmed_name = trimmed_name[1:]

        return trimmed_name.replace('/', '_')

    def get_module(self, template_env):
        if self.is_hierarchical:
            self.svgen_map = registry('SVGenMap')

            context = {
                'module_name': self.sv_module_name,
                'generics': [],
                'intfs': list(self.sv_port_configs()),
                'inst': [],
                'has_local_rst': self.has_local_rst
            }

            for child in self.node.local_interfaces():
                svgen = self.svgen_map[child]
                contents = svgen.get_inst(template_env)
                if contents:
                    context['inst'].append(contents)

            for child in self.node.local_modules():
                svgen = self.svgen_map[child]
                if hasattr(svgen, 'get_inst'):
                    contents = svgen.get_inst(template_env)
                    if contents:
                        context['inst'].append(contents)

            return template_env.render_local(__file__, "hier_module.j2",
                                             context)

    def update_port_name(self, port, name):
        port['name'] = name

    def get_inst(self, template_env):
        param_map = self.params

        in_port_map = [(port.basename, self.get_in_port_map_intf_name(port))
                       for port in self.node.in_ports]

        out_port_map = [(port.basename, self.get_out_port_map_intf_name(port))
                        for port in self.node.out_ports]

        rst_name = 'local_rst' if self.svgen_map[
            self.node.parent].has_local_rst else 'rst'

        try:
            if self.node.out_ports[-1].basename == 'rst_o':
                rst_name = 'rst'
        except:
            pass

        context = {
            'rst_name': rst_name,
            'module_name': self.sv_module_name,
            'inst_name': self.sv_inst_name,
            'param_map': param_map,
            'port_map': OrderedDict(in_port_map + out_port_map)
        }

        return template_env.snippets.module_inst(**context)


class SVTopGen(SVModuleGen):
    @property
    def sv_module_name(self):
        return "top"


class SVGenSVModPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace']['Gear'] = SVModuleGen
        cls.registry['SVGenModuleNamespace']['RTLNodeDesign'] = SVTopGen

        cls.registry['SVGenSystemVerilogPaths'].extend(
            [COMMON_SVLIB_DIR, COOKBOOK_SVLIB_DIR])



import socket
import logging
import argparse
import sys
import os
import time
import pexpect.fdpexpect


# def send_cmd(sock, cmd, log_level='INFO', timeout=30):
def send_cmd(sock, cmd, log_level='INFO'):

    log_level = getattr(logging, log_level)

    errs = []
    ret = []

    try:
        # pexp = pexpect.fdpexpect.fdspawn(sock)
        # pexp.sendline(cmd)
        sock.send((cmd.replace('\n', '\r') + '\r').encode())

        if not cmd.startswith('exit'):
            data = ''
            while (1):
                data += sock.recv(4096).decode()

                if data.endswith('% '):
                    data = data.replace('% ', '')
                    break

            for line in data.strip().split('\n'):
                # for line in pexp.before.decode().split('\n'):
                if line.startswith('ERROR:') or line.startswith('WARNING:'):
                    errs.append(line)
                else:
                    ret.append(line)

    except socket.timeout as e:
        print('Command timeout', file=sys.stderr)
    except socket.error as e:
        print(e, file=sys.stderr)

    sock.close()
    return ret, errs


def connect_and_send(sock,
                     port,
                     log_level,
                     timeout,
                     cmd,
                     return_err=False,
                     check=False):
    sock.connect(('', port))
    sock.settimeout(timeout)

    if not check:
        # ret, errs = send_cmd(sock, cmd, log_level, timeout)
        ret, errs = send_cmd(sock, cmd, log_level)
        print('\n'.join(ret))
        print('\n'.join(errs), file=sys.stderr)
        if errs and return_err:
            return -1


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(prog="GTKWave deamon TCP server")

    parser.add_argument(
        '-p', dest='port', type=int, default=60000, help="Server port")

    parser.add_argument(
        '-c',
        dest='check',
        action='store_true',
        help="Only check if server is up")

    parser.add_argument(
        '-l', dest='log_level', default='INFO', help="Logging level")

    parser.add_argument(
        '-t', dest='timeout', type=int, default=30, help="Timeout")

    parser.add_argument('cmd', metavar='cmd', help="GTKWave TCL command")

    parser.add_argument(
        '-e', '--error', action='store_true', help="Return error")

    args = parser.parse_args(argv[1:])

    client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    try:
        return connect_and_send(client_socket, args.port, args.log_level,
                                args.timeout, args.cmd, args.error, args.check)
    except socket.error as err:
        if args.check:
            return -1

    command = [
        'python',
        os.path.join(os.path.dirname(__file__), 'pywave_server.py'),
        '-p',
        # f'{args.port}', '-l', f'{args.log_level}'
        f'{args.port}',
        '-l',
        'DEBUG'
    ]

    pid = os.fork()
    if pid == 0:
        os.setsid()
        os.umask(0)
        os.execv(sys.executable, command)
    else:
        time.sleep(1)
        return connect_and_send(client_socket, args.port, args.log_level,
                                args.timeout, args.cmd, args.error)


if __name__ == '__main__':
    main(argv=[
        'pyvado', '-l', 'DEBUG', 'get_projects'
        # '{\nget_projects\nclose_projects\n}'
        # 'foreach {p} [get_projects] {\ncurrent_project p\n}'
        # 'version'
    ])
    # main()



from pygears import PluginBase, registry
from pygears.core.hier_node import HierVisitorBase, NamedHierNode, HierNode
from pygears.rtl.gear import RTLGearNodeGen, RTLNode
import inspect


class RTLNodeDesign(RTLNode):
    def __init__(self):
        super().__init__(None, '')


class GearHierRoot(NamedHierNode):
    def __init__(self, root):
        super().__init__('')
        self.in_ports = []
        self.out_ports = []
        self.root = root


class RTLNodeGearRoot(RTLGearNodeGen):
    def __init__(self, module):
        HierNode.__init__(self)
        self.node = RTLNodeDesign()
        self.gear = GearHierRoot(module)

        namespace = registry('SVGenModuleNamespace')
        self.node.params['svgen'] = {'svgen_cls': namespace['RTLNodeDesign']}
        self.module = module

        for p in self.gear.in_ports:
            self.node.add_in_port(p.basename, p.producer, p.consumer, p.dtype)

        for p in self.gear.out_ports:
            self.node.add_out_port(p.basename, p.producer, p.consumer, p.dtype)


class RTLNodeInstVisitor(HierVisitorBase):
    def __init__(self):
        self.cur_hier = None
        self.design = None
        self.namespace = registry('RTLGearGenNamespace')
        self.svgen_map = registry('RTLNodeMap')

    def NamedHierNode(self, module):
        self.design = RTLNodeGearRoot(module)
        self.cur_hier = self.design

    def instantiate(self, module):
        svgen = module.params.get('svgen', {})
        # node_cls = svgen.get('node_cls', None)
        if 'node_cls' in svgen:
            node_cls = svgen['node_cls']
        else:
            node_cls = self.namespace.get(module.definition, None)

            if node_cls is None:
                for base_class in inspect.getmro(module.__class__):
                    if base_class.__name__ in self.namespace:
                        node_cls = self.namespace[base_class.__name__]
                        break

        if node_cls:
            svgen_inst = node_cls(module, parent=self.cur_hier)
            self.svgen_map[module] = svgen_inst
        else:
            svgen_inst = None

        if self.cur_hier is None:
            self.design = svgen_inst

        return svgen_inst

    def Gear(self, module):
        inst = self.instantiate(module)

        if inst:
            self.cur_hier = inst
            super().HierNode(module)
            self.cur_hier = self.cur_hier.parent

        return True


def rtl_inst(top, conf):
    v = RTLNodeInstVisitor()
    v.visit(top)

    return v.design


class RTLNodeInstPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['RTLNodeNamespace'] = {}
        cls.registry['RTLGearGenNamespace'] = {
            'Gear': RTLGearNodeGen,
        }
        cls.registry['RTLNodeMap'] = {}

    @classmethod
    def reset(cls):
        cls.registry['RTLNodeMap'] = {}



import string
from collections import Counter


class HierYielderBase:
    def visit(self, node):
        import inspect
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                yield from getattr(self, base_class.__name__)(node)

    def HierNode(self, node):
        if hasattr(node, "child"):
            # Iterate over a copy in case visiting removes or adds children
            # from the parent
            for c in list(node.child):
                yield from self.visit(c)


class HierVisitorBase:
    def visit(self, node):
        import inspect
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                if getattr(self, base_class.__name__)(node):
                    return

    def HierNode(self, node):
        if hasattr(node, "child"):
            # Iterate over a copy in case visiting removes or adds children
            # from the parent
            for c in list(node.child):
                self.visit(c)


class HierNode:
    def __init__(self, parent=None):
        self.child = []
        self.parent = None
        if parent:
            parent.add_child(self)

    def add_child(self, module):
        self.child.append(module)
        module.parent = self

    def clear(self):
        for c in self.child.copy():
            if hasattr(c, 'remove'):
                c.remove()

        self.child = []

    def remove(self):
        if self.parent:
            self.parent.child.remove(self)

    def root(self):
        node = self
        while node.parent is not None:
            node = node.parent

        return node


def find_unique_names(names):
    def get_stem(s):
        return s.rstrip(string.digits)

    stems = list(map(get_stem, names))

    names_cnt = Counter(names)
    stems_cnt = Counter(stems)

    def unique(s):
        return names_cnt[s] <= 1

    def unique_stem(s):
        return stems_cnt[s] <= 1

    indexes = {k: 0 for k in stems_cnt}
    for name, stem in zip(names, stems):
        new_name = name
        if not unique_stem(stem):
            while ((new_name == stem) or (not unique(new_name))):
                names_cnt[new_name] -= 1
                new_name = f'{stem}{indexes[stem]}'
                if new_name in names_cnt:
                    names_cnt[new_name] += 1
                else:
                    names_cnt[new_name] = 1
                indexes[stem] += 1

        if new_name != name:
            yield new_name
        else:
            yield None


class NamedHierNode(HierNode):
    def __init__(self, basename=None, parent=None):
        super().__init__(parent)
        if basename is not None:
            self.basename = basename
            if parent:
                parent.unique_rename()

    def unique_rename(self):
        child_names = [c.basename for c in self.child]
        for child, new_name in zip(self.child, find_unique_names(child_names)):
            if new_name:
                child.basename = new_name

    def __getitem__(self, path):
        parts = path.split("/")

        for child in self.child:
            if hasattr(child, 'basename') and child.basename == parts[0]:
                break
        else:
            raise KeyError()

        if len(parts) == 1:
            return child
        else:
            return child["/".join(parts[1:])]

    @property
    def name(self):
        if self.parent:
            return '/'.join([self.parent.name, self.basename])
        else:
            return self.basename

    def is_descendent(self, node):
        return node.name.startswith(self.name)



from pygears.registry import PluginBase, bind, registry
from pygears.rtl.inst import rtl_inst
from pygears.rtl.connect import rtl_connect
from pygears.rtl.channel import RTLChannelVisitor, RTLOutChannelVisitor
from pygears.util.find import find
from .generate import svgen_generate
from .inst import svgen_inst


def svgen(top=None, **conf):

    if top is None:
        top = registry('HierRoot')
    elif isinstance(top, str):
        top = find(top)

    bind('SVGenConf', conf)
    for oper in registry('SVGenFlow'):
        top = oper(top, conf)

    return top


class SVGenPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['SVGenConf'] = {}
        cls.registry['SVGenFlow'] = [
            rtl_inst, rtl_connect, RTLChannelVisitor, RTLOutChannelVisitor, svgen_inst,
            svgen_generate
        ]



from pygears.typing import TypingNamespacePlugin, typeof
from pygears.typing import Int, Uint, Queue, Tuple, Union, Array


def cast(dtype, cast_type):
    if typeof(cast_type, Int) and (not cast_type.is_specified()):
        if typeof(dtype, Uint):
            return Int[int(dtype) + 1]
        elif typeof(dtype, Int):
            return dtype
        else:
            return Int[int(dtype)]
    if typeof(cast_type, Uint) and (not cast_type.is_specified()):
        if typeof(dtype, Int):
            return Uint[int(dtype) - 1]
        else:
            return Uint[int(dtype)]
    elif typeof(cast_type, Tuple) and (not cast_type.is_specified()):
        if typeof(dtype, Queue) or typeof(dtype, Union):
            return Tuple[dtype[0], dtype[1:]]
        elif typeof(dtype, Tuple):
            return dtype
        elif typeof(dtype, Array):
            return Tuple[(dtype[0], )*len(dtype)]
    elif (typeof(cast_type, Union) and typeof(dtype, Tuple)
          and len(dtype) == 2 and not cast_type.is_specified()):
        return Union[(dtype[0], )*(2**int(dtype[1]))]
    else:
        return cast_type


class CastTypePlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace']['cast'] = cast



import sys
import pdb
import fnmatch

fn_skip = ['<*', f'{sys.prefix}/lib/*']

module_skip = [
    'pygears.core.*', 'pygears.registry', 'pygears.util.*', 'inspect', 'enum',
    'pygears.typing.*', 'pygears.sim.*', '<*', 'pygears.common.const'
]


def is_skipped_frame(self, frame):
    fn = frame.f_code.co_filename
    mn = frame.f_globals.get('__name__')

    # print(f'Entering: {fn}: {mn}')

    if fn is not None:
        for pattern in fn_skip:
            if fnmatch.fnmatch(fn, pattern):
                return True

    if mn is not None:
        for pattern in module_skip:
            if fnmatch.fnmatch(mn, pattern):
                return True

    return False


def do_up(self, arg):
    frame_found = 0
    new_index = self.curindex
    count = int(arg or 1)

    while not frame_found == count:
        if new_index == 0:
            self.error('Oldest frame')
            break

        new_index -= 1

        if not is_skipped_frame(self, self.stack[new_index][0]):
            frame_found += 1

    if frame_found:
        self._select_frame(new_index)


def do_down(self, arg):
    frame_found = 0
    new_index = self.curindex
    count = int(arg or 1)

    while not frame_found == count:
        new_index += 1

        if new_index == len(self.stack):
            self.error('Oldest frame')
            break

        if not is_skipped_frame(self, self.stack[new_index][0]):
            frame_found += 1

    if frame_found:
        self._select_frame(new_index)


def stop_here(self, frame):
    if is_skipped_frame(self, frame):
        return False

    if frame is self.stopframe:
        if self.stoplineno == -1:
            return False
        return frame.f_lineno >= self.stoplineno

    if not self.stopframe:
        return True
    return False


def print_stack_trace(self):
    try:
        for frame_lineno in self.stack:
            if not is_skipped_frame(self, frame_lineno[0]):
                self.print_stack_entry(frame_lineno)
    except KeyboardInterrupt:
        pass


def patch_pdb():
    if sys.gettrace():
        p = sys.gettrace().__self__
        p.stop_here = stop_here.__get__(p, pdb.Pdb)
        p.do_up = do_up.__get__(p, pdb.Pdb)
        p.do_down = do_down.__get__(p, pdb.Pdb)
        p.do_u = do_up.__get__(p, pdb.Pdb)
        p.do_d = do_down.__get__(p, pdb.Pdb)
        p.print_stack_trace = print_stack_trace.__get__(p, pdb.Pdb)

    pdb.Pdb.stop_here = stop_here
    pdb.Pdb.do_up = do_up
    pdb.Pdb.do_down = do_down
    pdb.Pdb.do_u = do_up
    pdb.Pdb.do_d = do_down
    pdb.Pdb.print_stack_trace = print_stack_trace


def unpatch_pdb():
    pass



from pygears.core.gear import alternative, gear
from pygears.typing import Integer, Int, Uint
from pygears.core.intf import IntfOperPlugin
from pygears.util.hof import oper_reduce


def div_type(dtypes):
    length = int(dtypes[0])

    for i in range(1, len(dtypes)-1):
        length -= int(dtypes[i]) + 1

    if any(issubclass(d, Int) for d in dtypes):
        return Int[length]

    return Uint[length]


@gear(svgen={'svmod_fn': 'div.sv'}, enablement=b'len(din) == 2')
def div(*din: Integer, din0_signed=b'typeof(din0, Int)',
        din1_signed=b'typeof(din1, Int)') -> b'div_type(din)':
    pass


@alternative(div)
@gear
def div_vararg(*din: Integer, enablement=b'len(din) > 2') -> b'div_type(din)':
    return oper_reduce(din, div)


class DivIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__div__'] = div
        cls.registry['IntfOperNamespace']['__floordiv__'] = div
